STARTUP_TIMEOUT: float
"""How long to wait for the render server to start in seconds."""

RENDER_TIMEOUT: float
"""Timeout per rendering request in seconds."""

NODEJS_BINARY: str
"""nodejs binary to run javascript."""



from __future__ import annotations

import inspect
import sys
import textwrap
import warnings
from collections.abc import Generator
from contextlib import AbstractContextManager, contextmanager
from types import FunctionType, GenericAlias
from typing import Literal, LiteralString, Self

import dotenv
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined
from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, SettingsConfigDict

from ._utils import copy_func


def _type_str(cls: type, field: FieldInfo) -> str:
    if isinstance(field.annotation, GenericAlias) or not isinstance(field.annotation, type):
        return str(field.annotation)
    if field.annotation.__module__ in {"builtins", cls.__module__}:
        return field.annotation.__qualname__
    return f"{field.annotation.__module__}.{field.annotation.__qualname__}"


_docstring_template = """Allows users to customize settings for the `{package}` package.

Settings here will generally be for advanced use-cases and should be used with caution.

For setting an option use :func:`~{package}.{name}.override` (local) or set the attributes directly (global)
i.e., `{package}.{name}.my_setting = foo`. For assignment by environment variable, use the variable name in
all caps with `{env_prefix}` as the prefix before import of `{package}`.
"""


class Settings(BaseSettings):
    '''Base class for package settings.

    This class can be subclassed by individual packages to get package-specific settings handling.
    Settings will be validated on assignment thanks to Pydantic. The class requires the arguments
    `exported_object_name` and `docstring_style`, which will be used to construct a suitable
    docstring (see the examples).

    Both a settings instance and its `override` method should be added to the package documentation.

    Thanks to Pydantic Settings, settings values will also be loaded from environment variables or `.env`
    files. Environment variables must be prefixex with `$PACKAGE_NAME_` to take effect, where `$PACKAGE_NAME`
    is the name of the package of the subclass. This can be overridden by passing `env_prefix=CUSTOMPREFIX`
    as class argument.

    Examples:
        >>> from typing import Annotated
        ... from pydantic import Field
        ... from scverse_misc import Settings
        ...
        ...
        ... class MySettings(Settings, exported_object_name="settings", docstring_style="numpy"):
        ...     eps: Annotated[float, Field(gt=0, lt=1)] = 1e-8
        ...     """Small epsilon for numerical stability."""
        ...
        ...     use_optional_feature: bool = False
        ...     """Whether to use the optional feature."""
        ...
        ...
        ... settings = MySettings()
    '''

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return init_settings, env_settings, dotenv_settings

    @staticmethod
    def _get_packagename(subcls: type[Settings]) -> str:
        package_name = subcls.__module__
        dotidx = package_name.find(".")
        if dotidx > -1:
            package_name = package_name[:dotidx]
        return package_name

    def __init_subclass__(subcls, *, exported_object_name: str, docstring_style: Literal["google", "numpy", "scverse"]):
        if (config := subcls.__dict__.get("model_config")) is not None:
            if not config.get("validate_assignment", True):
                warnings.warn("`validate_assignment=False` is not supported, overriding.", RuntimeWarning, stacklevel=2)
            if not config.get("use_attribute_docstrings", True):
                warnings.warn(
                    "`use_attribute_docstrings=False` is not supported, overriding.", RuntimeWarning, stacklevel=2
                )
            if config.get("env_file") is not None:
                warnings.warn(
                    "Setting a custom env_file location is not supported, overriding.", RuntimeWarning, stacklevel=2
                )
        else:
            config = SettingsConfigDict()

        config["validate_assignment"] = True
        config["use_attribute_docstrings"] = True
        config["env_file"] = dotenv.find_dotenv()

        if not config.get("env_prefix"):
            config["env_prefix"] = f"{__class__._get_packagename(subcls)}_"  # type: ignore[name-defined] # https://github.com/python/mypy/issues/4177
        subcls.model_config = config

        super().__init_subclass__()

    @contextmanager
    def override(self, **overrides: object) -> Generator[None]:
        """Context manager for local setting overrides.

        Subclasses will get a version with a docstring detailing the available parameters.
        """
        oldsettings = {argname: getattr(self, argname) for argname in overrides.keys()}
        try:
            for argname, argval in overrides.items():
                setattr(self, argname, argval)
            yield
        finally:
            for argname, argval in reversed(oldsettings.items()):
                setattr(self, argname, argval)

    def reset(self, *names: LiteralString) -> AbstractContextManager[frozenset[LiteralString]]:
        """Reset passed settings to their default values.

        Can be used as a context manager to make the resets temporary.
        On `__enter__`, the context manager returns the settings that have been changed.
        """
        prev_values = {name: getattr(self, name) for name in names if name in self.model_fields_set}

        # since we want to allow using this method imperatively,
        # eagerly do the reset here instead of returning a context manager with a lazy `__enter__`.
        for name in prev_values:
            default = type(self).model_fields[name].get_default()
            setattr(self, name, default)
            self.model_fields_set.remove(name)

        class Cm(AbstractContextManager[frozenset[str]]):
            def __enter__(_self) -> frozenset[str]:
                return frozenset(prev_values)

            def __exit__(_self, *_: object) -> None:
                for arg, value in prev_values.items():
                    setattr(self, arg, value)

        return Cm()

    @classmethod
    def __pydantic_init_subclass__(  # type: ignore[override]
        subcls: type[Self], *, exported_object_name: str, docstring_style: Literal["google", "numpy", "scverse"]
    ) -> None:
        subcls.__doc__ = (
            _docstring_template.format(
                package=__class__._get_packagename(subcls),  # type: ignore[name-defined] # https://github.com/python/mypy/issues/4177
                name=exported_object_name,
                env_prefix=subcls.model_config["env_prefix"].upper(),
            )
            + "\n\nThe following options are available:\n"
        )
        override_doc = "Provides local override via keyword arguments as a context manager.\n\n"
        if docstring_style == "google":
            override_doc += "Args:\n"
        else:
            override_doc += "Parameters\n----------\n"
        for fname, field in subcls.model_fields.items():
            subcls.__doc__ += f"""
.. attribute:: {exported_object_name}.{fname}
   :type: {_type_str(subcls, field)}\n"""
            if field.default is not PydanticUndefined:
                subcls.__doc__ += f"   :value: {field.default!r}\n"

            description = ""
            if field.description is not None:
                subcls.__doc__ += f"\n{textwrap.indent(field.description, '   ')}\n"
                description += field.description

            if docstring_style == "google":
                override_doc += (
                    f"""    {fname} ({_type_str(subcls, field)}): {textwrap.indent(description, "        ")}\n"""
                )
            else:
                annot = "" if docstring_style == "scverse" else f" : {_type_str(subcls, field)}"
                override_doc += f"""\
{fname}{annot}
{textwrap.indent(description, "    ")}\n"""

        subcls.override = _copy_override(  # type: ignore[method-assign,type-var]
            subcls, subcls.override, override_doc, return_annotation=AbstractContextManager[None]
        )
        subcls.reset = _copy_reset(subcls, subcls.reset)  # type: ignore[method-assign,type-var]


class CustomRepr(str):
    def __repr__(self) -> str:
        return self


def _copy_override[F: FunctionType](cls: type[Settings], func: F, doc: str, return_annotation: object) -> F:
    from ._utils import Overrides

    parameters = [
        inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
        *[
            inspect.Parameter(
                n, inspect.Parameter.KEYWORD_ONLY, default=CustomRepr("<no change>"), annotation=f.annotation
            )
            for n, f in cls.model_fields.items()
        ],
    ]
    overrides = Overrides(
        __doc__=doc,
        __module__=cls.__module__,
        __qualname__=f"{cls.__qualname__}.{func.__name__}",
        __signature__=inspect.Signature(parameters, return_annotation=return_annotation),
        __annotations__={
            **{name: field.annotation for name, field in cls.model_fields.items()},
            "return": return_annotation,
        },
    )
    if sys.version_info >= (3, 14):
        from annotationlib import Format

        str_annotations = {n: _type_str(cls, f) for n, f in cls.model_fields.items()}
        overrides["__annotate__"] = lambda fmt: (
            overrides["__annotations__"] if fmt != Format.STRING else str_annotations
        )

    return copy_func(func, **overrides)


def _copy_reset[F: FunctionType](cls: type[Settings], func: F) -> F:
    from ._utils import Overrides

    args_t = Literal[tuple(cls.model_fields.keys())]  # type: ignore[valid-type]
    parameters = [
        inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY),
        inspect.Parameter("args", inspect.Parameter.VAR_POSITIONAL, annotation=args_t),
    ]
    return_annotation = AbstractContextManager[frozenset[args_t]]  # type: ignore[valid-type]
    overrides = Overrides(
        __module__=cls.__module__,
        __qualname__=f"{cls.__qualname__}.{func.__name__}",
        __signature__=inspect.Signature(parameters, return_annotation=return_annotation),
        __annotations__={"args": args_t, "return": return_annotation},
    )
    if sys.version_info >= (3, 14):
        from annotationlib import Format

        str_annotations = {n: str(t) for n, t in overrides["__annotations__"].items()}
        overrides["__annotate__"] = lambda fmt: (
            overrides["__annotations__"] if fmt != Format.STRING else str_annotations
        )

    return copy_func(func, **overrides)



import functools
import inspect
import sys
from collections.abc import Callable, Mapping
from functools import WRAPPER_ASSIGNMENTS
from types import FunctionType
from typing import ParamSpec, TypedDict, TypeVar, TypeVarTuple, Unpack, cast


class _BaseOverrides(TypedDict, total=False):
    __module__: str
    __name__: str
    __qualname__: str
    __doc__: str
    __signature__: inspect.Signature
    __annotations__: Mapping[str, object]
    __type_params__: tuple[TypeVar | TypeVarTuple | ParamSpec, ...]


if sys.version_info >= (3, 14):
    from annotationlib import Format

    class Overrides(_BaseOverrides, total=False):
        __annotate__: Callable[[Format], Mapping[str, object]]
else:

    class Overrides(_BaseOverrides, total=False):
        pass


def copy_func[F: FunctionType](func: F, /, **overrides: Unpack[Overrides]) -> F:
    kw = dict(kwdefaults=func.__kwdefaults__) if sys.version_info >= (3, 13) else {}
    new = FunctionType(
        func.__code__, func.__globals__, name=func.__name__, argdefs=func.__defaults__, closure=func.__closure__, **kw
    )
    for key, value in overrides.items():
        setattr(new, key, value)
    copy = set(WRAPPER_ASSIGNMENTS) - overrides.keys()
    wrapper = functools.update_wrapper(new, func, assigned=copy)
    del wrapper.__wrapped__  # otherwise sphinx will try to document that.
    return cast("F", wrapper)



from __future__ import annotations

import inspect
import sys
from contextlib import suppress
from functools import wraps
from textwrap import indent
from typing import TYPE_CHECKING, LiteralString
from warnings import warn

if sys.version_info >= (3, 13):
    from warnings import deprecated as _deprecated
else:
    from typing_extensions import deprecated as _deprecated

if TYPE_CHECKING:
    from collections.abc import Callable


__all__ = ["deprecated", "deprecated_arg", "Deprecation"]


class Deprecation(str):
    """Utility class storing information on deprecated functionality.

    Args:
        version_deprecated: The version of the package where the functionality was deprecated.
        msg: The deprecation message.
    """

    version_deprecated: LiteralString

    def __new__(cls, version_deprecated: LiteralString, msg: LiteralString = "") -> LiteralString:  # type: ignore[misc]  # typing.Intersection doesn’t exist yet
        if not msg:
            msg = ""  # be lenient here, people don’t want to see “None” or “False” here
        obj = super().__new__(cls, msg)
        obj.version_deprecated = version_deprecated
        return obj


def _deprecated_at[F: Callable[..., object]](
    msg: Deprecation, *, category: type[Warning] = FutureWarning, stacklevel: int = 1
) -> Callable[[F], F]:
    """Decorator to indicate that a class, function, or overload is deprecated.

    Wraps :func:`warnings.deprecated` and additionally modifies the docstring to include a deprecation notice.

    Args:
        msg: The deprecation message.
        category: The category of the warning that will be emitted at runtime.
        stacklevel: The stack level of the warning.

    Examples:
        >>> @deprecated(Deprecation("0.2", "Use bar() instead."))
        ... def foo(baz):
        ...     pass
    """

    def decorate(func: F) -> F:
        kind = "function" if func.__name__ == func.__qualname__ else "method"
        warnmsg = f"The {kind} {func.__name__} is deprecated and will be removed in the future"

        doc = inspect.getdoc(func)
        docmsg = f".. version-deprecated:: {msg.version_deprecated}"
        if len(msg):
            docmsg += f"\n{indent(msg, 3 * ' ')}"
            warnmsg += f". {msg}" if msg.count("\n") == 0 else f":\n{indent(msg, 4 * ' ')}"
        else:
            warnmsg += "."

        if doc is None:
            doc = docmsg
        else:
            lines = doc.splitlines()
            body = "\n".join(lines[1:])
            doc = f"{lines[0]}\n\n{docmsg}\n{body}"
        func.__doc__ = doc

        return _deprecated(warnmsg, category=category, stacklevel=stacklevel)(func)

    return decorate


if TYPE_CHECKING:
    deprecated = _deprecated
else:
    deprecated = _deprecated_at


def deprecated_arg[**P, R](
    arg: LiteralString, msg: Deprecation, *, category: type[Warning] = FutureWarning, stacklevel: int = 1
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to indicate that a function argument is deprecated.

    Emits a warning when the decorated function is called with the deprecated argument and addtionally modifies the
    docstring to include a deprecation notice.

    Args:
        arg: The deprecated argument.
        msg: The deprecation message.
        category: The category of the warning that will be emitted at runtime.
        stacklevel: The stack level of the warning.

    Examples:
        >>> @deprecated_arg("bar", Deprecation("0.2", "The functionality has moved to the baz() function."))
        ... def foo(baz, bar=1):
        ...     pass
    """

    def decorate(func: Callable[P, R]) -> Callable[P, R]:
        warnmsg = f"The argument {arg} is deprecated and will be removed in the future."
        if len(msg):
            warnmsg += f" {msg}"

        if func.__doc__ is not None:
            with suppress(ImportError):
                func.__doc__ = _deprecate_arg_doc(func.__doc__, arg=arg, msg=msg)

        sig = inspect.signature(func)
        param = sig.parameters[arg]

        @wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if (
                param.kind in (inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                and arg in kwargs
            ):
                warn(warnmsg, category=category, stacklevel=stacklevel + 1)
            else:
                bound = sig.bind(*args, **kwargs)
                if arg in bound.arguments and bound.arguments[arg] != param.default:
                    warn(warnmsg, category=category, stacklevel=stacklevel + 1)

            return func(*args, **kwargs)

        return wrapped

    return decorate


def _deprecate_arg_doc(doc: str, *, arg: str, msg: Deprecation) -> str:
    from pydocstring import Docstring, Section, SectionKind, Style, emit_google, emit_numpy, parse

    docmsg = f".. version-deprecated:: {msg.version_deprecated}"
    if len(msg):
        docmsg += f"\n   {msg}"

    parsed = parse(doc)
    if parsed.style is Style.PLAIN:
        return doc

    model = parsed.to_model()
    if found := next(
        (
            (s, section, p, par)
            for s, section in enumerate(model.sections)
            if section.kind in {SectionKind.PARAMETERS, SectionKind.KEYWORD_PARAMETERS, SectionKind.OTHER_PARAMETERS}
            for p, par in enumerate(section.parameters)
            if arg in par.names
        ),
        None,
    ):
        s, section, p, par = found
        if par.description is not None:
            docmsg += f"\n\n{par.description}"
        par.description = docmsg
        params = list(section.parameters)
        params[p] = par
        sections = list(model.sections)
        sections[s] = Section(section.kind, parameters=params)
        model = Docstring(
            summary=model.summary,
            extended_summary=model.extended_summary,
            deprecation=model.deprecation,
            sections=sections,
        )
    match parsed.style:
        case Style.GOOGLE:
            return emit_google(model)
        case Style.NUMPY:
            return emit_numpy(model)
        case _:  # pragma: no cover
            raise AssertionError



from __future__ import annotations

import inspect
import sys
from contextlib import nullcontext
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Literal, cast, get_args

import pytest
from pydantic import Field, ValidationError
from pydantic.fields import FieldInfo
from pydantic_settings import SettingsConfigDict
from sphinx.application import Sphinx
from sphinx.ext.napoleon import GoogleDocstring, NumpyDocstring  # type: ignore[attr-defined]

from scverse_misc import Settings

if TYPE_CHECKING:
    # Static version of the class returned by the `settings_class` fixture
    class DummySettings(Settings, exported_object_name="settings"):
        field_bool: bool = False
        field_no_docstring: int = 42
        field_int_range: int = 1


pytest_plugins = ["sphinx.testing.fixtures"]


@pytest.fixture
def docstring_style(request: pytest.FixtureRequest) -> Literal["google", "numpy", "scverse"]:
    return getattr(request, "param", "google")


@pytest.fixture
def settings_class(docstring_style: Literal["google", "numpy", "scverse"]) -> type[DummySettings]:
    class _DummySettings(Settings, exported_object_name="settings", docstring_style=docstring_style):
        field_bool: bool = False
        """Boolean field."""

        field_no_docstring: int = 42

        field_int_range: Annotated[int, Field(ge=0, le=4)] = 1
        """Integer range field."""

    return cast("type[DummySettings]", _DummySettings)


@pytest.fixture
def settings(settings_class: type[DummySettings]) -> DummySettings:
    return settings_class()


def test_defaults_override() -> None:
    with (
        pytest.warns(RuntimeWarning, match="validate_assignment=False"),
        pytest.warns(RuntimeWarning, match="use_attribute_docstrings=False"),
        pytest.warns(RuntimeWarning, match="custom env_file location"),
    ):

        class WarnSettings(Settings, exported_object_name="settings", docstring_style="google"):
            model_config = SettingsConfigDict(
                validate_assignment=False, use_attribute_docstrings=False, env_file="mydotenv"
            )

            field_bool: bool = False

    settings = WarnSettings()
    with pytest.raises(ValidationError):
        settings.field_bool = 2  # type: ignore[assignment]


@pytest.mark.parametrize("v", [2, 4])
def test_env_vars(monkeypatch: pytest.MonkeyPatch, settings_class: type[DummySettings], v: int) -> None:
    """Test that the env var prefix is derived from the module name."""
    monkeypatch.setenv("TESTS_FIELD_INT_RANGE", str(v))
    settings = settings_class()
    assert settings.field_int_range == v


def test_validate_assignment(settings: DummySettings) -> None:
    with pytest.raises(ValidationError):
        settings.field_bool = 2  # type: ignore[assignment]
    with pytest.raises(ValidationError):
        settings.field_int_range = -1


def test_override(settings: DummySettings) -> None:
    with settings.override(field_bool=True):
        assert settings.field_bool is True
    assert settings.field_bool is False


def test_override_error(settings: DummySettings) -> None:
    with pytest.raises(ValidationError):
        with settings.override(field_int_range=3, field_no_docstring=1.1):
            pass
    assert settings.field_no_docstring == 42
    assert settings.field_int_range == 1


@pytest.mark.parametrize("temp", [True, False], ids=["temporary", "permanent"])
def test_reset(settings: DummySettings, temp: bool) -> None:
    default = settings.field_bool
    settings.field_bool = not default
    undo_reset = settings.reset("field_bool")
    with undo_reset if temp else nullcontext():
        assert settings.field_bool is default
    assert settings.field_bool is (not default if temp else default)


def test_reset_signature(settings: DummySettings) -> None:
    sig = inspect.signature(settings.reset)
    assert get_args(sig.parameters["args"].annotation) == ("field_bool", "field_no_docstring", "field_int_range")


@pytest.mark.skipif(sys.version_info < (3, 14), reason="requires annotationlib")
def test_reset_annotations(settings: DummySettings) -> None:
    from contextlib import AbstractContextManager

    import annotationlib

    assert annotationlib.get_annotations(settings.reset) == {
        "args": Literal["field_bool", "field_no_docstring", "field_int_range"],
        "return": AbstractContextManager[frozenset[Literal["field_bool", "field_no_docstring", "field_int_range"]]],
    }


@pytest.mark.parametrize("docstring_style", ["google", "numpy", "scverse"], indirect=True)
def test_docs(docstring_style: Literal["google", "numpy"], settings: DummySettings) -> None:
    parser = GoogleDocstring if docstring_style == "google" else NumpyDocstring
    lines = parser(inspect.getdoc(settings) or "").lines()

    assert lines[0].endswith("`tests` package.")

    current_field: FieldInfo | None = None
    field_iter = iter(type(settings).model_fields.items())
    for line in lines:
        if line.startswith(".. attribute::"):
            current_field_name, current_field = next(field_iter)
            assert line.endswith(current_field_name)
        elif current_field is not None:
            line = line.strip()
            if line.startswith(":type:"):
                assert current_field.annotation is not None
                assert line.endswith(current_field.annotation.__name__)
            elif line.startswith(":value:"):
                assert line.endswith(repr(current_field.default))
            elif len(line) > 0 and current_field.description is not None:
                assert line == current_field.description


@pytest.mark.parametrize("docstring_style", ["google", "numpy", "scverse"], indirect=True)
def test_override_docs(docstring_style: Literal["google", "numpy"], settings: DummySettings) -> None:
    parser = GoogleDocstring if docstring_style == "google" else NumpyDocstring
    lines = parser(inspect.getdoc(settings.override) or "").lines()

    current_field: FieldInfo | None = None
    field_iter = iter(type(settings).model_fields.items())
    for line in lines:
        if line.startswith(":param"):
            current_field_name, current_field = next(field_iter)
            description = f" {current_field.description}" if current_field.description is not None else ""
            # no default here, as the default is “leave this value alone”
            assert line.startswith(f":param {current_field_name}:{description}")
        elif current_field is not None and len(line) > 0:
            assert current_field.annotation is not None
            assert line == f":type {current_field_name}: {current_field.annotation.__name__}"


@pytest.mark.parametrize(
    ("attr", "expected"),
    [
        pytest.param("string", "str", id="builtin"),
        pytest.param("path", "pathlib.Path", id="3rd-party"),
        # same module as `S`, so no leading `tests.test_settings.`
        pytest.param("local", "test_annotation_format.<locals>.Local", id="same-module"),
    ],
)
def test_annotation_format(attr: str, expected: str) -> None:
    """Test that annotation references work correctly."""

    class Local: ...

    class S(Settings, exported_object_name="s", docstring_style="google"):
        if attr == "string":
            string: str
        if attr == "path":
            path: Path
        if attr == "local":
            local: Local

    lines = (inspect.getdoc(S) or "").splitlines()
    lines = lines[lines.index(f".. attribute:: s.{attr}") + 1 :]

    assert lines == [f"   :type: {expected}"]


@pytest.fixture(scope="session", autouse=True)
def _sphinx_config(sphinx_test_tempdir: Path) -> None:
    """Since we only need one, we use this instead of static roots like `@pytest.mark.sphinx('html', testroot="mybook")`."""
    p = sphinx_test_tempdir / "root" / "conf.py"
    p.parent.mkdir(parents=True)
    p.write_text("""
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon", "sphinx_autodoc_typehints"]
typehints_defaults = "braces"
""")


@pytest.mark.parametrize("docstring_style", ["scverse"])
@pytest.mark.parametrize("parent", ["class", "object"])
def test_sphinx_autodoc_typehints(
    subtests: pytest.Subtests,
    app: Sphinx,
    settings_class: type[DummySettings],
    settings: DummySettings,
    parent: Literal["class", "object"],
) -> None:
    import sphinx_autodoc_typehints
    from sphinx.ext.napoleon import NumpyDocstring  # type: ignore[attr-defined]

    obj = (settings if parent == "object" else settings_class).override
    lines = (inspect.getdoc(obj) or "").splitlines()
    lines = NumpyDocstring(lines, app.config, app, "method", "", obj).lines()

    with subtests.test("napoleon"):
        # test that napoleon can parse things correctly
        # especially the last parameter could fail to parse if there are not enough trailing newlines
        for name in settings_class.model_fields:
            assert f":param {name}:" in "\n".join(lines)

    sphinx_autodoc_typehints.process_docstring(app, "method", "", obj, options=None, lines=lines)

    with subtests.test("type"):  # no need to test all parameters
        assert (
            r":type field_bool: :sphinx_autodoc_typehints_type:`\:py\:class\:\`bool\`` (default: ``<no change>``)"
            in lines
        )
    with subtests.test("rtype"):
        assert (
            r":rtype: :sphinx_autodoc_typehints_type:`\:py\:class\:\`\~contextlib.AbstractContextManager\`\\ \\\[\:py\:obj\:\`None\`\]`"
            in lines
        )



import inspect
import warnings
from collections.abc import Callable
from typing import Literal, cast, get_args

import pytest
from sphinx.ext.napoleon import GoogleDocstring, NumpyDocstring  # type: ignore[attr-defined]

from scverse_misc import Deprecation, deprecated, deprecated_arg


@pytest.fixture(
    params=[
        pytest.param(None, id="no_message"),
        pytest.param("Test message.", id="short_message"),
        pytest.param("Test\nmessage.", id="long_message"),
    ]
)
def msg(request: pytest.FixtureRequest) -> str | None:
    return cast(str | None, request.param)


type DocstringStyles = Literal["no_docstring", "short", "long_numpystyle", "long_googlestyle"]


@pytest.fixture(params=get_args(DocstringStyles.__value__))
def docstring_style(request: pytest.FixtureRequest) -> DocstringStyles:
    return cast(DocstringStyles, request.param)


@pytest.fixture
def docstring(docstring_style: DocstringStyles) -> str | None:
    match docstring_style:
        case "no_docstring":
            return None
        case "short":
            return "Test function"
        case "long_numpystyle":
            return """Test function

            This is a test.

            Parameters
            ----------
            positional_only_no_default
                foo
            positional_only_default
                bar
            positional_or_keyword_default
                baz
            keyword_only_default
                foobar
            """
        case "long_googlestyle":
            return """Test function

            This is a test.

            Args:
                positional_only_no_default: foo
                positional_only_default: bar lorem ipsum
                    test
                positional_or_keyword_default: baz
                keyword_only_default: foobar
            """


@pytest.fixture
def func(msg: str | None, docstring: str | None) -> Callable[..., int]:
    def _func(
        positional_only_no_default: int,
        positional_only_default: int = 1337,
        /,
        positional_or_keyword_default: int = 42,
        *,
        keyword_only_default: float = 3.1415,
    ) -> int:
        return 42

    _func.__doc__ = docstring
    return _func


@pytest.fixture
def deprecated_func(msg: str | None, func: Callable[..., int]) -> Callable[..., int]:
    return deprecated(Deprecation("foo", msg or ""))(func)


def test_deprecation_decorator(deprecated_func: Callable[..., int], docstring: str | None, msg: str | None) -> None:
    with pytest.warns(FutureWarning, match="deprecated"):
        assert deprecated_func(1, 2) == 42

    assert deprecated_func.__doc__ is not None
    lines = deprecated_func.__doc__.expandtabs().splitlines()
    offset = 0 if docstring is None else 2

    if docstring is not None:
        lines_orig = docstring.expandtabs().splitlines()
        assert lines[0] == lines_orig[0]
        assert len(lines[1].strip()) == 0, "expected empty line following summary"

    assert lines[offset].startswith(".. version-deprecated")
    if msg is None:
        assert len(lines) == offset + 1 or not lines[offset + 1].startswith("   ")
    else:
        msg_lines = msg.splitlines()
        msg_indented = [f"   {line}" for line in msg_lines]
        assert lines[offset + 1 : offset + 1 + len(msg_lines)] == msg_indented


@pytest.mark.parametrize(
    "arg",
    ("positional_only_no_default", "positional_only_default", "positional_or_keyword_default", "keyword_only_default"),
)
def test_deprecated_arg_decorator(
    func: Callable[..., int], msg: str | None, arg: str, docstring_style: DocstringStyles
) -> None:
    deprecated_func = deprecated_arg(arg, Deprecation("2.718", msg or ""))(func)
    with pytest.warns(FutureWarning, match=f"{arg} is deprecated"):
        assert deprecated_func(1, 2, 3, keyword_only_default=4.0) == 42

    if arg != "positional_only_no_default":
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            assert deprecated_func(1) == 42

    parser: type[NumpyDocstring] | type[GoogleDocstring] | None = None
    if docstring_style == "long_numpystyle":
        parser = NumpyDocstring
    elif docstring_style == "long_googlestyle":
        parser = GoogleDocstring

    if parser is None:
        return

    lines = parser(inspect.getdoc(deprecated_func) or "").lines()

    for i, line in enumerate(lines):
        if line.startswith(prefix := f":param {arg}: "):
            prefixlen = len(prefix)
            if msg is not None:
                stripped = lines[i + 1].strip()
                assert stripped == ".. version-deprecated:: 2.718"
                assert lines[i + 2][prefixlen:] == f"   {msg}"
                assert not lines[i + 3]
                assert lines[i + 4][:prefixlen] == " " * prefixlen
            else:
                assert line == f":param {arg}: .. version-deprecated:: 2.718"
                assert not lines[i + 1]
                assert lines[i + 2][:prefixlen] == " " * prefixlen



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

import pytest

from scverse_misc import _extensions as extensions
from scverse_misc import make_register_namespace_decorator

if TYPE_CHECKING:
    from collections.abc import Generator


class Greeter(Protocol):
    def __init__(self, obj: DummyClass) -> None: ...
    def greet(self) -> str: ...


class DummyClass:
    foo: list[object] = []
    bar = None

    @property
    def baz(self) -> None: ...
    def foobar(self) -> None: ...

    dummy: Greeter  # available when using `dummy_namespace` fixture


register_dummy_namespace = make_register_namespace_decorator(
    DummyClass, "obj", "register_dummy_namespace", docstring_style="google"
)


@pytest.fixture
def obj() -> DummyClass:
    """Create a basic object for testing."""
    return DummyClass()


@pytest.fixture(autouse=True)
def _cleanup_dummy() -> Generator[None, None, None]:
    """Automatically cleanup dummy namespace after each test."""
    original = getattr(DummyClass, "dummy", None)
    yield
    if original is not None:
        DummyClass.dummy = original
    elif hasattr(DummyClass, "dummy"):
        delattr(DummyClass, "dummy")


@pytest.fixture
def dummy_namespace() -> type:
    """Create a basic dummy namespace class."""

    @register_dummy_namespace("dummy")
    class DummyNamespace:
        def __init__(self, obj: DummyClass) -> None:
            self._obj = obj

        def greet(self) -> str:
            return "hello"

    return DummyNamespace


def test_accessor_namespace() -> None:
    """Test the behavior of the AccessorNameSpace descriptor.

    This test verifies that:
    - When accessed at the class level (i.e., without an instance), the descriptor
      returns the namespace type.
    - When accessed via an instance, the descriptor instantiates the namespace,
      passing the instance to its constructor.
    - The instantiated namespace is then cached on the instance such that subsequent
      accesses of the same attribute return the cached namespace instance.
    """

    # Define a dummy namespace class to be used via the descriptor.
    class DummyNamespace:
        def __init__(self, obj: Dummy):
            self._obj = obj

        def foo(self) -> str:
            return "foo"

    class Dummy:
        dummy: DummyNamespace  # just typing, runtime added below

    descriptor: extensions.AccessorNameSpace[Dummy, DummyNamespace] = extensions.AccessorNameSpace(
        "dummy", DummyNamespace
    )

    # When accessed on the class, it should return the namespace type.
    ns_class = descriptor.__get__(None, Dummy)
    assert ns_class is DummyNamespace

    # When accessed via an instance, it should instantiate DummyNamespace.
    dummy_obj = Dummy()
    ns_instance = descriptor.__get__(dummy_obj, Dummy)
    assert isinstance(ns_instance, DummyNamespace)
    assert ns_instance._obj is dummy_obj

    # __get__ should cache the namespace instance on the object.
    # Subsequent access should return the same cached instance.
    assert dummy_obj.dummy is ns_instance


def test_descriptor_instance_caching(dummy_namespace: type, obj: DummyClass) -> None:
    """Test that namespace instances are cached on individual DummyClass objects."""
    # First access creates the instance
    ns_instance = obj.dummy
    # Subsequent accesses should return the same instance
    assert obj.dummy is ns_instance


def test_register_namespace_basic(dummy_namespace: type, obj: DummyClass) -> None:
    """Test basic namespace registration and access."""
    assert obj.dummy.greet() == "hello"


def test_register_namespace_override(dummy_namespace: type) -> None:
    """Test namespace registration and override behavior."""
    assert hasattr(DummyClass, "dummy")

    # Override should warn and update the namespace
    with pytest.warns(UserWarning, match="Overriding existing custom namespace 'dummy'"):

        @register_dummy_namespace("dummy")
        class DummyNamespaceOverride:
            def __init__(self, obj: DummyClass) -> None:
                self._obj = obj

            def greet(self) -> str:
                return "world"

    # Verify the override worked
    obj = DummyClass()
    assert obj.dummy.greet() == "world"


@pytest.mark.parametrize(
    "attr",
    [
        "foo",
        "bar",
        "baz",
        "foobar",
    ],
)
def test_register_existing_attributes(attr: str) -> None:
    """
    Test that registering an accessor with a name that is a reserved attribute of DummyClass raises an attribute error.
    """
    with pytest.raises(AttributeError, match=f"cannot override reserved attribute {attr!r}"):

        @register_dummy_namespace(attr)
        class DummyNamespace:
            def __init__(self, obj: DummyClass) -> None:
                self._obj = obj


def test_valid_signature() -> None:
    """Test that a namespace with valid signature is accepted."""

    @register_dummy_namespace("valid")
    class ValidNamespace:
        def __init__(self, obj: DummyClass) -> None:
            self.obj = obj


def test_missing_param() -> None:
    """Test that a namespace missing the second parameter is rejected."""
    with pytest.raises(
        TypeError, match=r"Namespace initializer must accept a DummyClass instance as the second parameter\."
    ):

        @register_dummy_namespace("missing_param")
        class MissingParamNamespace:
            def __init__(self) -> None:
                pass


def test_wrong_name() -> None:
    """Test that a namespace with wrong parameter name is rejected."""
    with pytest.raises(
        TypeError, match=r"Namespace initializer's second parameter must be named 'obj', got 'notobj'\."
    ):

        @register_dummy_namespace("wrong_name")
        class WrongNameNamespace:
            def __init__(self, notobj: DummyClass) -> None:
                self.notobj = notobj


def test_wrong_annotation() -> None:
    """Test that a namespace with wrong parameter annotation is rejected."""
    with pytest.raises(
        TypeError,
        match=r"Namespace initializer's second parameter must be annotated as the 'DummyClass' class, got 'int'\.",
    ):

        @register_dummy_namespace("wrong_annotation")
        class WrongAnnotationNamespace:
            def __init__(self, obj: int) -> None:
                self.obj = obj


def test_missing_annotation() -> None:
    """Test that a namespace with missing parameter annotation is rejected."""
    with pytest.raises(AttributeError):

        @register_dummy_namespace("missing_annotation")
        class MissingAnnotationNamespace:
            def __init__(self, obj) -> None:  # type: ignore[no-untyped-def]
                self.obj = obj


def test_both_wrong() -> None:
    """Test that a namespace with both wrong name and annotation is rejected."""
    with pytest.raises(
        TypeError,
        match=(
            r"Namespace initializer's second parameter must be named 'obj', got 'info'\. "
            r"And must be annotated as 'DummyClass', got 'str'\."
        ),
    ):

        @register_dummy_namespace("both_wrong")
        class BothWrongNamespace:
            def __init__(self, info: str) -> None:
                self.info = info



from contextlib import suppress

from ._deprecated import Deprecation, deprecated, deprecated_arg
from ._extensions import ExtensionNamespace, make_register_namespace_decorator

__all__ = ["ExtensionNamespace", "make_register_namespace_decorator", "deprecated", "deprecated_arg", "Deprecation"]

with suppress(ImportError):
    from ._settings import Settings

    __all__.append("Settings")



"""System to add extension attributes to classes.

Based off of the extension framework in Polars:
https://github.com/pola-rs/polars/blob/main/py-polars/polars/api.py
"""

from __future__ import annotations

import inspect
import sys
import warnings
from itertools import islice
from typing import TYPE_CHECKING, Literal, Protocol, get_type_hints, overload, runtime_checkable

if TYPE_CHECKING:
    from collections.abc import Callable, Set


__all__ = ["make_register_namespace_decorator", "ExtensionNamespace"]


@runtime_checkable
class ExtensionNamespace(Protocol):
    """Protocol for extension namespaces.

    Enforces that the namespace initializer accepts a class with the proper `__init__` method.
    Protocol's can't enforce that the `__init__` accepts the correct types. See
    `_check_namespace_signature` for that. This is mainly useful for static type
    checking with mypy and IDEs.
    """

    def __init__(self, instance: object) -> None:
        """Used to enforce the correct signature for extension namespaces."""


class AccessorNameSpace[T, NameSpT: ExtensionNamespace]:
    """Establish property-like namespace object for user-defined functionality."""

    def __init__(self, name: str, namespace: type[NameSpT]) -> None:
        self._accessor = name
        self._ns = namespace

    @overload
    def __get__(self, instance: None, cls: type[T]) -> type[NameSpT]: ...

    @overload
    def __get__(self, instance: T, cls: type[T]) -> NameSpT: ...

    def __get__(self, instance: T | None, cls: type[T]) -> NameSpT | type[NameSpT]:
        if instance is None:
            return self._ns

        ns_instance = self._ns(instance)
        setattr(instance, self._accessor, ns_instance)
        return ns_instance


def _check_namespace_signature(ns_class: type, cls: type, canonical_instance_name: str) -> None:
    """Validate the signature of a namespace class for extensions.

    This function ensures that any class intended to be used as an extension namespace
    has a properly formatted `__init__` method such that:

    1. Accepts at least two parameters (self and the instance of the extended class)
    2. Has `canonical_instance_name` as the name of the second parameter
    3. Has the second parameter properly type-annotated as `ns_class` or any equivalent import alias

    The function performs runtime validation of these requirements before a namespace
    can be registered through the `register_namespace` decorator.

    Args:
        ns_class: The namespace class to validate.
        cls: The class that is being extended.
        canonical_instance_name: The name of the `ns_class` constructor argument.

    Raises:
        TypeError: If the `__init__` method has fewer than 2 parameters (missing the instance parameter).
        AttributeError: If the second parameter of `__init__` lacks a type annotation.
        TypeError: If the second parameter of `__init__` is not named `canonical_instance_name`.
        TypeError: If the second parameter of `__init__` is not annotated as the `ns_class` class.
        TypeError: If both the name and type annotation of the second parameter are incorrect.

    """
    sig = inspect.signature(ns_class.__init__)  # type: ignore[misc]  # https://github.com/python/mypy/issues/21236
    params = sig.parameters

    # Ensure there are at least two parameters (self and mdata)
    if len(params) < 2:
        raise TypeError(f"Namespace initializer must accept a {cls.__name__} instance as the second parameter.")

    # Get the second parameter (expected to be `canonical_instance_name`)
    [_, param, *_] = params.values()
    if param.annotation is inspect.Parameter.empty:
        raise AttributeError(
            f"Namespace initializer's second parameter must be annotated as the {cls.__name__!r} class, got empty annotation."
        )

    name_ok = param.name == canonical_instance_name

    # Resolve the annotation using get_type_hints to handle forward references and aliases.
    try:
        type_hints = get_type_hints(ns_class.__init__)  # type: ignore[misc]  # https://github.com/python/mypy/issues/21236
        resolved_type = type_hints.get(param.name, param.annotation)
    except NameError as e:
        raise NameError(
            f"Namespace initializer's second parameter must be named {canonical_instance_name!r}, got '{param.name}'."
        ) from e

    type_ok = resolved_type is cls

    match (name_ok, type_ok):
        case (True, True):
            return  # Signature is correct.
        case (False, True):
            raise TypeError(
                f"Namespace initializer's second parameter must be named {canonical_instance_name!r}, got {param.name!r}."
            )
        case (True, False):
            type_repr = getattr(resolved_type, "__name__", str(resolved_type))
            raise TypeError(
                f"Namespace initializer's second parameter must be annotated as the {cls.__name__!r} class, got {type_repr!r}."
            )
        case _:
            type_repr = getattr(resolved_type, "__name__", str(resolved_type))
            raise TypeError(
                f"Namespace initializer's second parameter must be named {canonical_instance_name!r}, got {param.name!r}. "
                f"And must be annotated as {cls.__name__!r}, got {type_repr!r}."
            )


def _create_namespace[NameSpT: ExtensionNamespace](
    name: str, cls: type, reserved_namespaces: Set[str], canonical_instance_name: str
) -> Callable[[type[NameSpT]], type[NameSpT]]:
    """Register custom namespace against the underlying class."""

    def namespace(ns_class: type[NameSpT]) -> type[NameSpT]:
        _check_namespace_signature(ns_class, cls, canonical_instance_name)  # Perform the runtime signature check
        if name in reserved_namespaces:
            raise AttributeError(f"cannot override reserved attribute {name!r}")
        elif hasattr(cls, name):
            warnings.warn(
                f"Overriding existing custom namespace {name!r} (on {cls.__name__!r})", UserWarning, stacklevel=2
            )
        setattr(cls, name, AccessorNameSpace(name, ns_class))
        return ns_class

    return namespace


def _indent_string_lines(string: str, indentation_level: int, skip_lines: int = 0) -> str:
    minspace = sys.maxsize
    for line in islice(string.splitlines(), 1, None):
        for i, char in enumerate(line):
            if not char.isspace():
                minspace = min(minspace, i)
                break
    if minspace == sys.maxsize:  # single-line string
        minspace = 0
    return "\n".join(
        " " * 4 * indentation_level + sline if i >= skip_lines else sline
        for i, line in enumerate(string.splitlines())
        if (sline := (line[minspace:] if i > 0 else line)) or True
    )


def make_register_namespace_decorator[NameSpT: ExtensionNamespace](
    cls: type, canonical_instance_name: str, decorator_name: str, docstring_style: Literal["google", "numpy", "scverse"]
) -> Callable[[str], Callable[[type[NameSpT]], type[NameSpT]]]:
    """Create a decorator for registering custom functionality with a class.

    The decorator will allow your users to extend `cls` objects with custom methods and properties
    organized under a namespace. The namespace becomes accessible as an attribute on `cls` instances,
    providing a clean way for users to add domain-specific functionality without modifying the `cls`
    class itself.

    The return decorator will have a docstring describing how to use it along with examples.

    Args:
        cls: The class to be made extensible.
        canonical_instance_name: The typical name of an instance of `cls`, e.g. `adata` for `AnnData`. This
            is used for run-time checking of constructor signatures of the namespace classes.
        decorator_name: The name under which the decorator is accessible in your package. This is used for
            the examples in the decorator docstring.
        docstring_style: Whether the docstring of the generated decorator should conform to `"numpy"` or
            `"google"` style. We also support variant of `"numpy"` called `"scverse"`,
            which does not duplicate type annotation in docstrings.
    """
    # Reserved namespaces include accessors built into cls and all current attributes of cls
    reserved_namespaces = set(dir(cls))

    def decorator(name: str) -> Callable[[type[NameSpT]], type[NameSpT]]:
        return _create_namespace(name, cls, reserved_namespaces, canonical_instance_name)

    decorator_arg_description = f"""Name under which the accessor should be registered. This will be the attribute name
            used to access your namespace's functionality on {cls.__name__} objects (e.g., `instance.name`).
            Cannot conflict with existing {cls.__name__} attributes. The list of reserved attributes includes
            everything outputted by `dir({cls.__name__})`."""
    decorator_return_description = "A decorator that registers the decorated class as a custom namespace."
    decorator_notes = f"""Implementation requirements:

        1. The decorated class must have an `__init__` method that accepts exactly one parameter
           (besides `self`) named `{canonical_instance_name}` and annotated with type :class:`~{cls.__module__}.{cls.__name__}`.
        2. The namespace will be initialized with the {cls.__name__} object on first access and then
           cached on the instance.
        3. If the namespace name conflicts with an existing namespace, a warning is issued.
        4. If the namespace name conflicts with a built-in {cls.__name__} attribute, an AttributeError is raised."""
    decorator_examples = f""">>> @{decorator_name}("do_something")
        ... class DoSomething:
        ...     def __init__(self, {canonical_instance_name}: {cls.__name__}):
        ...         self._obj = {canonical_instance_name}
        ...
        ...     def has_foo(self) -> bool:
        ...         return hasattr(self._obj, "foo")
        >>>
        >>> # Create a {cls.__name__} object
        >>> obj = {cls.__name__}()
        >>>
        >>> # use the registered namespace
        >>> obj.do_something.has_foo()
        False"""

    decorator.__doc__ = f"""Decorator for registering custom functionality with a :class:`~{cls.__module__}.{cls.__name__}` object.

    This decorator allows you to extend {cls.__name__} objects with custom methods and properties
    organized under a namespace. The namespace becomes accessible as an attribute on {cls.__name__}
    instances, providing a clean way to you to add domain-specific functionality without modifying
    the {cls.__name__} class itself, or extending the class with additional methods as you see fit in your workflow.
    """

    if docstring_style == "google":
        decorator.__doc__ += f"""
    Args:
        name: {_indent_string_lines(decorator_arg_description, 3, 1)}

    Returns:
        {_indent_string_lines(decorator_return_description, 2, 1)}

    Notes:
        {_indent_string_lines(decorator_notes, 2, 1)}

    Examples:
        {_indent_string_lines(decorator_examples, 2, 1)}
    """
    else:
        decorator.__doc__ += f"""
    Parameters
    ----------
    name
        {_indent_string_lines(decorator_arg_description, 2, 1)}

    Returns
    -------
    {_indent_string_lines(decorator_return_description, 1, 1)}

    Notes
    -----
    {_indent_string_lines(decorator_notes, 1, 1)}

    Examples
    --------
    {_indent_string_lines(decorator_examples, 1, 1)}
    """

    return decorator
